/*
 * NicheStack TCP/IP stack initialization and Operating System Start in main()
 * for UDP offload example. 
 * 
 */
  
#include <stdio.h>

/* MicroC/OS-II definitions */
#include "includes.h"

/* Nichestack definitions */
#include "ipport.h"
#include "libport.h"
#include "osport.h"

/* UDP offload demo definitions */
#include "demo_control.h"
#include "demo_tasks.h"

#define INITIALIZATION_TASK_PRIORITY 28
#define TASK_STACKSIZE 2048

/* 
 * Definition of task stack for the initial task which will initialize the NicheStack
 * TCP/IP Stack and then initialize the rest of the example tasks. 
 */
OS_STK    InitializationTaskStk[TASK_STACKSIZE];

void InitializationTask(void *task_data)
{
  INT8U error_code;
  
  // initialize the OS Statistics task so it will collect CPU usage for us
  OSStatInit();
  
  /*
   * Initialize Altera NicheStack TCP/IP Stack - Nios II Edition specific code.
   * NicheStack is initialized from a task, so that RTOS will have started, and 
   * I/O drivers are available.  Two tasks are created:
   *    "Inet main"  task with priority 2
   *    "clock tick" task with priority 3
   */   
  alt_iniche_init();
  netmain(); 

  /* 
   * Wait for the network stack to be ready before proceeding. 
   * iniche_net_ready indicates that TCP/IP stack is ready, and IP address is obtained.
   */
  while (!iniche_net_ready)
    TK_SLEEP(1);

  /* Now that the stack is running, perform the application initialization steps */
  
  /* Application Specific Task Launching Code Block Begin */

  printf("\nStack is running...\n\n");

  // start the client and server tasks required to facilitate the UDP offload 
  // initiation and termination over the network.
  TK_NEWTASK(&the_stream_server_task);
  TK_NEWTASK(&the_stream_client_task);

  // initialize our InterNiche menu commnads so they can be invoked from a
  // a console or telnet session.
  udp_demo_init();

  /* Application Specific Task Launching Code Block End */
  
  /*This task is deleted because there is no need for it to run again */
  error_code = OSTaskDel(OS_PRIO_SELF);
  if(error_code){
	  printf("Error deleting initialization task...\n");
  }
  
  while (1); /* Correct Program Flow should never get here */
}

/* 
 * Main creates a single task, InitializationTask, and starts task scheduler.
 */

int main (int argc, char* argv[], char* envp[])
{
  
  INT8U error_code;

  /* Clear the RTOS timer */
  OSTimeSet(0);

  /* InitializationTask will initialize the NicheStack
   * TCP/IP Stack and then initialize the rest of the example 
   * RTOS structures and tasks. 
   */  
  error_code = OSTaskCreateExt(InitializationTask,
                             NULL,
                             (void *)&InitializationTaskStk[TASK_STACKSIZE],
                             INITIALIZATION_TASK_PRIORITY,
                             INITIALIZATION_TASK_PRIORITY,
                             InitializationTaskStk,
                             TASK_STACKSIZE,
                             NULL,
                             0);
  if(error_code){
	  printf("Error creating initialization task...\n");
	  while(1);
  }

  /*
   * As with all MicroC/OS-II designs, once the initial thread(s) and 
   * associated RTOS resources are declared, we start the RTOS. That's it!
   */
  OSStart();

  
  while(1); /* Correct Program Flow never gets here. */

  return -1;
}
