//
//	This file contains functions that collect and report CPU utilization
//	statistics from the uCOS-II RTOS.  There are many dependencies within these
//	functions for various pieces of the InterNiche TCP/IP stack as well.  The
//	InterNiche dependencies could be easily removed to retarget an environment
//	that did not utilize the InterNiche stack.
//	
//	User must have an altera_avalon_timer peripheral in their system to
//	dedicate to this function.  There are two macros that must be set in
//	my_os_hooks.h, that are required to identify your altera_avalon_timer to
//	these functions.
//
//	You must also properly initialize the uCOS-II statistics task by executing
//	OSStatInit() in your program.
//	
//	To avoid linking issues with the BSP implmentation of these functions, you
//	should add the __attribute__((weak)) to the corresponding function
//	definitions in os_cpu_c.c like this:
//	
//	void OSTaskSwHook (void) __attribute__((weak));
//	void OSTaskStatHook (void) __attribute__((weak));
//	void OSTimeTickHook (void) __attribute__((weak));
//

/* MicroC/OS-II definitions */
#include "includes.h"

#include "ipport.h"
#include "my_os_hooks.h"
#include "altera_avalon_timer_regs.h"

extern void cticks_hook(void);

//
//	OSTaskStatHook storage
//	
alt_u32 my_cpu_stats[CPU_STATS_COUNT];
alt_u32 my_current_cpu_stat = 0;

//
//	OSTimeTickHook storage
//	
alt_u32 TimeTickHook_running = 0;
alt_u32 longest_tick = 0;
alt_u32 lastTickHookTimer = 0;

//
//	OSTaskSwHook storage
//	
alt_u32 exitCount[OS_LOWEST_PRIO + 1] = { 0 };
alt_u32 runTime[OS_LOWEST_PRIO + 1] = { 0 };
alt_u32 maxTime[OS_LOWEST_PRIO + 1] = { 0 };
alt_u32 last_exitCount[OS_LOWEST_PRIO + 1] = { 0 };
alt_u32 last_runTime[OS_LOWEST_PRIO + 1] = { 0 };
alt_u32 last_maxTime[OS_LOWEST_PRIO + 1] = { 0 };
alt_u32 lastStatTimer = 0;
alt_u32 OSTaskSwHook_running = 0;

//
//	This routine simply stores the OSCPUUsage value calculated by OSTaskStat()
//	into a circular buffer of CPU_STATS_COUNT entries.  This information is 
//	displayed by the cpu_stats function.  To initialize the OSTaskStat()
//	function you must first call OSStatInit() to initialize it.  Be warned that
//	the default distribution of the OS has OS_TASK_STAT_STK_CHK_EN defined and
//	these stack checks can consume a considerable amount of time.
//	

//
//	This routine simply captures the value of the stat timer at each tick and
//	measures the amount of time since the last tick.  If the elapsed time
//	for the last tick was greater than the value in longest_tick, then we
//	update it with the new longest tick time.  You would expect that the
//	longest time between ticks should be at least greater than the tick
//	duration, but not much greater.  If some task or interrupt handler is
//	grossly disabling interrupts, you could spot that with a long tick duration.
//	The value of longest_tick is displayed with the cpu_stats function.
//	

//
//	This routine simply accumulates statistics on task switching from the RTOS.
//	The routine stores a count of the execution occurances of each task as well
//	as a running accumulation of execution time, as well as the maximum single
//	execution duration.  The statistics are accumulated into a current and a
//	last collection of data, where the amount of data in each of these
//	collections is determined by the frequency of the stat timer peripheral.
//	The current collection of statistic is copied into the last collection at
//	the point where the stat timer rolls over from 0x00000000 to 0xFFFFFFFF.
//	The values of these statistics can be displayed with the cpu_stats function.
//	

//
//	This routine displays the values collected by the three OS hook routines.
//	
int cpu_stats(void * pio) {

	OS_CPU_SR  cpu_sr = 0;
	int i;
    
	alt_u32 local_cpu_stats[CPU_STATS_COUNT];
	int max, min, ave;

	alt_u32 local_longest_tick;

    alt_u32 local_exitCount[OS_LOWEST_PRIO + 1];
    alt_u32 local_runTime[OS_LOWEST_PRIO + 1];
    alt_u32 local_maxTime[OS_LOWEST_PRIO + 1];
    alt_u32 local_last_exitCount[OS_LOWEST_PRIO + 1];
    alt_u32 local_last_runTime[OS_LOWEST_PRIO + 1];
    alt_u32 local_last_maxTime[OS_LOWEST_PRIO + 1];

    OS_ENTER_CRITICAL();
    
    // grab all the data
    
    for(i = 0 ; i < CPU_STATS_COUNT ; i++) {
		local_cpu_stats[i] = my_cpu_stats[i];
	}
    
    for(i = 0 ; i < OS_LOWEST_PRIO + 1 ; i++) {
    	local_exitCount[i] = exitCount[i];
    	local_runTime[i] = runTime[i];
    	local_maxTime[i] = maxTime[i];
    	local_last_exitCount[i] = last_exitCount[i];
    	local_last_runTime[i] = last_runTime[i];
    	local_last_maxTime[i] = last_maxTime[i];
    }
    
    local_longest_tick = longest_tick;
    
    OS_EXIT_CRITICAL();

    // process the data a bit
    max = 0;
    min = 100;
    ave = 0;
    for(i = 0 ; i < CPU_STATS_COUNT ; i++) {
    	
    	ave += local_cpu_stats[i];
    	
    	if(local_cpu_stats[i] > max) {
    		max = local_cpu_stats[i];
    	}

    	if(local_cpu_stats[i] < min) {
    		min = local_cpu_stats[i];
    	}
    }
    ave /= CPU_STATS_COUNT;
    
    // display the data
    ns_printf(pio,"\nCPU Statistics for last %u seconds:\n\n   MAX = %d%%   MIN = %d%%   AVE = %d%%\n\n", STATS_SECONDS, max, min, ave);
    ns_printf(pio,"MAX and MIN are for a single 1/%u second period\nAVE is for the entire %u second period\n\n", STATS_PER_SECOND, STATS_SECONDS);

    ns_printf(pio,"Longest Tick Spacing = %uus\n\n", local_longest_tick / (THE_STAT_TIMER_FREQ / 1000000));
    
    ns_printf(pio,"RECENT TASK STATISTICS - times are in microseconds, current data becomes last\n");
    ns_printf(pio,"                         data every %u seconds, clear_cpu_stats will reset\n\n", (0xFFFFFFFF / THE_STAT_TIMER_FREQ));
    ns_printf(pio,"   Prio | Exit Count   Run Time   Max Time |  Exit Last   Run Last   Max Last\n");
    for(i = 0 ; i < OS_LOWEST_PRIO + 1 ; i++) {
    	ns_printf(pio,"%7d |%11u%11u%11u |%11u%11u%11u\n",
    			i,
    			local_exitCount[i],
    			local_runTime[i] / (THE_STAT_TIMER_FREQ / 1000000), 
    			local_maxTime[i] / (THE_STAT_TIMER_FREQ / 1000000),
    			local_last_exitCount[i],
    			local_last_runTime[i] / (THE_STAT_TIMER_FREQ / 1000000), 
    			local_last_maxTime[i] / (THE_STAT_TIMER_FREQ / 1000000) );
    }
    ns_printf(pio,"\n");
    
	return 0;
}

//
//	This routine clears all the CPU statistics.
//	
int clear_cpu_stats(void * pio) {
	
	int i;
    OS_CPU_SR  cpu_sr = 0;
    
    OS_ENTER_CRITICAL();
    
    for(i = 0 ; i < CPU_STATS_COUNT ; i++) {
    	my_cpu_stats[i] = 0;
	}
    
    for(i = 0 ; i < OS_LOWEST_PRIO + 1 ; i++) {
    	exitCount[i] = 0;
    	runTime[i] = 0;
    	maxTime[i] = 0;
    	last_exitCount[i] = 0;
    	last_runTime[i] = 0;
    	last_maxTime[i] = 0;
    }
    
    longest_tick = 0;
    
    OS_EXIT_CRITICAL();

    ns_printf(pio,"CPU stats cleared...\n");

    return 0;
}

//
//	This routine gets the current value of the statistics timer.
//	
alt_u32 getStatTick(void) {
	
	alt_u32 statTimerLo;
	alt_u32 statTimerHi;
	alt_u32 thisStatTimer;
    OS_CPU_SR  cpu_sr = 0;
	
    OS_ENTER_CRITICAL();

	// grab the stat timer snapshot
	IOWR_ALTERA_AVALON_TIMER_SNAPL(THE_STAT_TIMER_BASE, 0x0000);
	statTimerLo = IORD_ALTERA_AVALON_TIMER_SNAPL(THE_STAT_TIMER_BASE) & 0xFFFF;
	statTimerHi = IORD_ALTERA_AVALON_TIMER_SNAPH(THE_STAT_TIMER_BASE) << 16;
	thisStatTimer = statTimerHi | statTimerLo;
    
    OS_EXIT_CRITICAL();
    
    return(thisStatTimer);
}

//
//	This routine initializes the statistics timer.
//	
void initStatTick(void) {
	
    OS_CPU_SR  cpu_sr = 0;
	
    OS_ENTER_CRITICAL();

	// set period to maximum
	IOWR_ALTERA_AVALON_TIMER_PERIODH(THE_STAT_TIMER_BASE, 0xFFFF);
	IOWR_ALTERA_AVALON_TIMER_PERIODL(THE_STAT_TIMER_BASE, 0xFFFF);
	
	// start the timer to run continuously with no interrupt
	IOWR_ALTERA_AVALON_TIMER_CONTROL(THE_STAT_TIMER_BASE, (ALTERA_AVALON_TIMER_CONTROL_START_MSK | ALTERA_AVALON_TIMER_CONTROL_CONT_MSK));

	OS_EXIT_CRITICAL();
}
